#!/bin/sh

## Replacement for heavyweight (and non-cross-platform) groupadd function
## To be used in fakeroot environment for handling groupadd calls during
## exportroot creation

# Author: Phil Estes <estesp@linux.vnet.ibm.com>

_PROGNAME=groupadd
_PROGVERSION=0.9.9

# group format:
# groupname:x:gid:[memberlist]  where [memberlist] is optional comma-separated list of usernames

# example call we must handle
# groupadd -r -o ldap -g 25

_REQUIRE_UNIQUE=1
_SYSTEM_GRP=0
_GROUP=/etc/group
_SYSMIN=100
_USRMIN=1000

function assert_arg()
{
    [ $# -lt 1 ] && {
        echo "$FUNCNAME: wrong number of args ($#) supplied"; return -1;
    }

    #make local copies of arguments
    local switch=$1;
    local arg=$2;

    [ -z "$arg" ] && {
        echo "missing argument to $switch"; return -1;
    }

    [ "${arg:0:1}" == "-" ] && {
        echo "missing argument to $switch"; return -1;
    }

    return 0;
}

function help()
{
    #make local copies of global variables
    local progname=$_PROGNAME; [ -z "$progname" ] && return -1;

    if [ $# -eq 0 ]; then
cat <<-EOM
'$progname' is a shell replacement for the groupadd binary

$progname [-ro] [-g <gid>] groupname

The following flags are supported in the shell replacement:
  -r        : create system group
  -o        : uniqueness not required
  -g <gid>  : requested gid

EOM
    fi
    return 0;
}

function parse_group()
{
    local oldIFS=$IFS;
    local tmpEntry;
    
    IFS="
"
    for line in `cat $_GROUP`; do
      IFS=:
      tmpEntry=( $line );
      IFS="
" 
      _GIDMAP[${tmpEntry[2]}]=${tmpEntry[0]};
      _GROUPS=( "${_GROUPS[@]}" "${tmpEntry[0]}" );
    done
    IFS=$oldIFS;
    return 0;
}

function create_group()
{
    [ $# -ne 2 ] && {
      echo "FAIL: create_group: bad arguments ($#)"; return -1;
    }
    local gid=$1;
    local groupname=$2;
    local min=$_USRMIN;

    parse_group || return -1;

    #assign an available GID if necessary

    [ "$gid" -eq 0 ] && {
       #we need to find the next available GID
       [ $_SYSTEM_GRP -eq 1 ] && {
         min=$_SYSMIN;
       }
       let i=$min;
       while [ -n "${_GIDMAP[$i]}" ]; do
         let "i+=1";
       done

       gid=$i
    } || {
      # if uniqueness required, verify GID not used by someone else
      [ $_REQUIRE_UNIQUE -eq 1 ] && {
        [ -n "${_GIDMAP[$gid]}" ] && {
          echo "FATAL: GID $gid already in use by groupname ${_GIDMAP[$gid]}"; return -1;
        }
      }
    }

    #the groupname must be unique, no matter if -o is passed or not
    for name in ${_GROUPS[@]}; do
      [[ "$name" == "$groupname" ]] && {
        echo "FATAL: Groupname $groupname already in use in $_GROUP"; return -1;
      }
    done

    #we should now have all pieces of information to be able to create the group in
    #$_GROUP
    echo "$groupname:x:$gid:" >>$_GROUP
    return 0;
}

function main()
{
    [ $# -eq 0 ] && { _PROGNAME=`basename $0` help; return $?; }

    local group;
    local gid;

    #parse the command line arguments
    while [ $# -gt 0 ]; do
        case $1 in
        -\? | --help)
            help --help; return $?;
            ;;
        -V | --version)
            echo "$_PROGNAME $_PROGVERSION"; return 0;
            ;;
        -o)
            _REQUIRE_UNIQUE=0; shift;
            ;;
        -r)
            _SYSTEM_GRP=1; shift;
            ;;
        -g)
            shift;
	    assert_arg '-g' $1 || return -1
            gid=$1; shift;
            ;;
        -*)
            echo "bailing out, unrecognized switch: $1"; return -1;
            ;;
        *)
	    [ -n "$group" ] && {
              echo "WARN: Previously specified groupname $group replaced with $1";
            }
            group=$1; shift;
            ;;
        esac
    done

    [ -z "$group" ] && {
      echo "Groupname must be specified on the command line"; return -1;
    }

    #we must be root to be able to perform the actual group creation steps..check now
    [[ `id -u` -ne 0 ]] && {
      echo "FATAL: You must be root to add groups"; return -1;
    }

    #for args that were specified, verify them to be reasonable

    [ -z "$gid" ] && {
      # user requests we find an unused gid
      gid=0;
    }

    [ -n "${gid//[[:digit:]]/}" ] && {
      #gid must be numeric
      echo "FATAL: Group ID must be a number ($gid)"; return -1;
    }

    create_group ${gid} "${group}";
    return $?;
}

export LC_ALL=C;
main "$@";
ret=$?;
exit $ret;
