#!/bin/sh

## Replacement for heavyweight (and non-cross-platform) useradd function
## To be used in fakeroot environment for handling useradd calls during
## exportroot creation

# Author: Phil Estes <estesp@linux.vnet.ibm.com>

_PROGNAME=useradd
_PROGVERSION=0.9.9

# passwd format:
# username:x:uid:gid:comment:homedir:shell

# example call we must handle
# useradd -r -o -g bin -u 66 -s /bin/false -c "Wnn System Account" -d /var/lib/wnn wnn

_REQUIRE_UNIQUE=1
_SYSTEM_ID=0
_DEFAULT_GROUP=100
_DEFAULT_SHELL=/bin/bash
_DEFAULT_HOME=/home
_PASSWD=/etc/passwd
_GROUP=/etc/group
_SYSMIN=100
_USRMIN=1000

function assert_arg()
{
    [ $# -lt 1 ] && {
        echo "$FUNCNAME: wrong number of args ($#) supplied"; return -1;
    }

    #make local copies of arguments
    local switch=$1;
    local arg=$2;

    [ -z "$arg" ] && {
        echo "missing argument to $switch"; return -1;
    }

    [ "${arg:0:1}" == "-" ] && {
        echo "missing argument to $switch"; return -1;
    }

    return 0;
}

function help()
{
    #make local copies of global variables
    local progname=$_PROGNAME; [ -z "$progname" ] && return -1;

    if [ $# -eq 0 ]; then
cat <<-EOM
'$progname' is a shell replacement for the useradd binary

$progname [-ro] [-g <gid>] [-u <uid>] [-s <shell>] [-c <comment>]
                [-d <homedir] username

The following flags are supported in the shell replacement:
  -r        : create system account
  -o        : uniqueness not required
  -g <gid>  : groupname or gid
  -u <uid>  : uid
  -s <shell>: shell
  -c <comm> : comment/name of user
  -d <home> : home dir setting

EOM
    fi
    return 0;
}

function parse_passwd()
{
    local oldIFS=$IFS;
    local tmpEntry;
    
    IFS="
"
    for line in `cat $_PASSWD`; do
      IFS=:
      tmpEntry=( $line );
      IFS="
" 
      _UIDMAP[${tmpEntry[2]}]=${tmpEntry[0]};
      _USERS=( "${_USERS[@]}" "${tmpEntry[0]}" );
    done
    IFS=$oldIFS;
    return 0;
}

function create_user()
{
    [ $# -ne 6 ] && {
      echo "FAIL: create_user: bad arguments ($#)"; return -1;
    }
    local uid=$1;
    local gid=$2;
    local username=$3;
    local homedir=$4;
    local shell=$5;
    local comment=$6;
    local min=$_USRMIN;

    parse_passwd || return -1;

    #assign an available UID if necessary

    [ "$uid" -eq 0 ] && {
       #we need to find the next available UID
       [ $_SYSTEM_ID -eq 1 ] && {
         min=$_SYSMIN;
       }
       let i=$min;
       while [ -n "${_UIDMAP[$i]}" ]; do
         let "i+=1";
       done

       uid=$i
    } || {
      # if uniqueness required, verify UID not used by someone else
      [ $_REQUIRE_UNIQUE -eq 1 ] && {
        [ -n "${_UIDMAP[$uid]}" ] && {
          echo "FATAL: UID $uid already in use by username ${_UIDMAP[$uid]}"; return -1;
        }
      }
    }

    #the username must be unique, no matter if -o is passed or not
    for name in ${_USERS[@]}; do
      [[ "$name" == "$username" ]] && {
        echo "FATAL: Username $username already in use in $_PASSWD"; return -1;
      }
    done

    #verify group ID is sane
    local ginfo=`grep ":$gid:" $_GROUP`;
    [ -z "$ginfo" ] && {
       echo "FATAL: Group ID $gid is not a valid group identifier"; return -1;
    }

    #we should now have all pieces of information to be able to create the user in
    #$_PASSWD..
    echo "$username:x:$uid:$gid:$comment:$homedir:$shell" >>$_PASSWD
    return 0;
}

function main()
{
    [ $# -eq 0 ] && { _PROGNAME=`basename $0` help; return $?; }

    local group;
    local gid;
    local uid;
    local username;
    local comment;
    local homedir;
    local shellname;

    #parse the command line arguments
    while [ $# -gt 0 ]; do
        case $1 in
        -\? | --help)
            help --help; return $?;
            ;;
        -V | --version)
            echo "$_PROGNAME $_PROGVERSION"; return 0;
            ;;
        -o)
            _REQUIRE_UNIQUE=0; shift;
            ;;
        -r)
            _SYSTEM_ID=1; shift;
            ;;
        -s)
            shift;
	    assert_arg '-s' $1 || return -1
            shellname=$1; shift;
            ;;
        -g)
            shift;
	    assert_arg '-g' $1 || return -1
            group=$1; shift;
            ;;
        -u)
            shift;
	    assert_arg '-u' $1 || return -1
            uid=$1; shift;
            ;;
        -c)
            shift;
	    assert_arg '-c' $1 || return -1
            comment=$1; shift;
            ;;
        -d)
            shift;
	    assert_arg '-d' $1 || return -1
            homedir=$1; shift;
            ;;
        -*)
            echo "bailing out, unrecognized switch: $1"; return -1;
            ;;
        *)
	    [ -n "$username" ] && {
              echo "WARN: Previously specified username $username replaced with $1";
            }
            username=$1; shift;
            ;;
        esac
    done

    [ -z "$username" ] && {
      echo "Username must be specified on the command line"; return -1;
    }
    #we must be root to be able to perform the actual user creation steps..check now
    [[ `id -u` -ne 0 ]] && {
      echo "FATAL: You must be root to add users"; return -1;
    }

    #for args that were specified, verify them to be reasonable
    # is group specified, if so, id or groupname
    [ -z "${group}" ] && {
      #group not specified..need to default
      group=$_DEFAULT_GROUP;
      gid=$group
    }

    [ -n "${group//[[:digit:]]/}" ] && {
      #group is a string; we must look it up in /etc/group
      gid=`grep "^$group:" $_GROUP | awk -F: '{ print $3 }'`;
      [ -z "$gid" ] && {
        echo "FATAL: unable to find group $group in $_GROUP"; return -1;
      }
    }

    [ -z "$gid" ] && {
      # this means we didn't pass through "group unset" and "group is string"..so we
      # got a numeric gid from the user..pass it along
      gid=$group;
    }

    #uid, if specified, must be a number
    [ -n "${uid//[[:digit:]]/}" ] && {
      echo "FATAL: uid must be a number ($uid)"; return -1;
    }
    
    #default rest of params that are unset:
    [ -z "$uid" ] && {
      uid=0;
    }

    [ -z "$shellname" ] && {
      shellname=$_DEFAULT_SHELL;
    }

    [ -z "$comment" ] && {
      comment="$username";
    }
    
    [ -z "$homedir" ] && {
      homedir=$_DEFAULT_HOME/$username;
    }

    create_user ${uid} ${gid} "${username}" "${homedir}" "${shellname}" "${comment}";
    return $?;
}

export LC_ALL=C;
main "$@";
ret=$?;
exit $ret;
